"""
 Author: kapil thangavelu <kvthan@wm.edu>

 Purpose: Provide for Intersect, Union, and Join
          Operations on ZSQL Results. Heavily
          based on gadfly, which contains the
          secret of the universe (relational anyways ;-)

 Created: 6/14/2002

 CopyRight: kapil thangavelu <kvthan@wm.edu> @2002
 
 License: X11 (see LICENSE.txt for full details)

 CVS: $Id: KjResults.py,v 1.2 2002/08/03 01:14:51 ender Exp $

"""
import unittest
from operator import add, sub, mul

from Shared.DC.ZRDB.Results import Results

from KjRelation import relation
from util import make_col, dump_rowset


try:
    from kjbuckets import *
except:
    from Products.ZGadflyDA.gadfly.kjbuckets0 import *


def relation_from_results(result):
    """
    convert a results object to a kj relation
    """
    return relation(tuple(result.names()),
                    result._data,
                    result._parent)


def results_from_relation(relation):
    """
    convert a kj relation to a results object
    """
    return Results((make_col(relation.schemeseq),
                    dump_rowset(relation)),
                    parent=relation.parent)

def union_results(*res):
    return results_from_relation(reduce(add, map(relation_from_results, res)))
    
def difference_results(*res):
    return results_from_relation(reduce(sub, map(relation_from_results, res)))
    
def join_results(*res):
    return results_from_relation(reduce(mul, map(relation_from_results, res)))


class ResultsTestCases(unittest.TestCase):

    from ExtensionClass import Base

    col1 = ('name', 'color', 'food')
    set1 = [('mike', 'blue', 'pizza'),
            ('john', 'green', 'yams'),
            ('suzan', 'red', 'pasta')]

    col1 = make_col(col1)    
    
    col2 = ('name', 'edu', 'state', 'age')
    set2 = [('mike', 'college', 'va', 21),
            ('john', 'high', 'ny', 22),
            ('suzan', 'grad', 'ut', 26),
            ('joesmoe', 'kinder', 'za', 102)]
    
    col2 = make_col(col2)

    col3 = ('name', 'color', 'food')
    set3 = [('loki', 'white', 'choco'),
            ('romulus', 'yellow', 'zima'),
            ('orion', 'blue', 'sushi'),
            ('john', 'green', 'yams'),
            ('suzan', 'red', 'pasta')]            

    col3 = make_col(col3)

    class p(Base):
        def __init__(self, v):
            self.v = v
        def __str__(self): return str(self.v)
        
    p1 = p(1)
    p2 = p(2)
    p3 = p(3)
    
    r1 = Results( (col1, set1), parent=p1 )
    r2 = Results( (col2, set2), parent=p2 )
    r3 = Results( (col3, set3), parent=p3 )
    

    def testJoin(self):
        r = join_results(self.r1,self.r2)

        assert len(r) == 3, "wrong number of join elements"

        s = self.r1.names()+self.r2.names()
        s.reverse()
        s.pop(s.index('name'))
        s.sort()
        
        indices = r.names()
        indices.sort()

        self.assertEqual(indices,
                         s,
                         "bad scheme for join")
        
    def testUnion(self):

        r = union_results(self.r1, self.r3)

        assert len(r) == 6, "wrong number of union elements"
        assert len(r.names()) == 3, "bad schema for union"
        
    def testDifference(self):

        r = difference_results(self.r1, self.r3)

        assert len(r) == 1, "wrong number of difference elements"
        assert len(r.names()) == 3, "bad schema for difference"        

if __name__ == '__main__':
    unittest.main()

